=== modified file 'bzrlib/branch.py'
--- bzrlib/branch.py	2005-12-19 20:31:01 +0000
+++ bzrlib/branch.py	2006-01-18 13:45:16 +0000
@@ -487,6 +487,27 @@
     def store_revision_signature(self, gpg_strategy, plaintext, revision_id):
         raise NotImplementedError('store_revision_signature is abstract')
 
+    def fileid_involved_between_revs(self, from_revid, to_revid):
+        """ This function returns the file_id(s) involved in the
+            changes between the from_revid revision and the to_revid
+            revision
+        """
+        raise NotImplementedError('fileid_involved_between_revs is abstract')
+
+    def fileid_involved(self, last_revid=None):
+        """ This function returns the file_id(s) involved in the
+            changes up to the revision last_revid
+            If no parametr is passed, then all file_id[s] present in the
+            repository are returned
+        """
+        raise NotImplementedError('fileid_involved is abstract')
+
+    def fileid_involved_by_set(self, changes):
+        """ This function returns the file_id(s) involved in the
+            changes present in the set 'changes'
+        """
+        raise NotImplementedError('fileid_involved_by_set is abstract')
+
 class BzrBranch(Branch):
     """A branch stored in the actual filesystem.
 
@@ -1115,6 +1136,67 @@
         self.revision_store.add(StringIO(gpg_strategy.sign(plaintext)), 
                                 revision_id, "sig")
 
+    def fileid_involved_between_revs(self, from_revid, to_revid):
+        """ This function returns the file_id(s) involved in the
+            changes between the from_revid revision and the to_revid
+            revision
+        """
+        w = self._get_inventory_weave( )
+        from_set = set(w.inclusions([w.lookup(from_revid)]))
+        to_set = set(w.inclusions([w.lookup(to_revid)]))
+        included = to_set.difference(from_set)
+        changed = map(w.idx_to_name,included)
+        return self._fileid_involved_by_set(changed)
+
+    def fileid_involved(self, last_revid=None):
+        """ This function returns the file_id(s) involved in the
+            changes up to the revision last_revid
+            If no parametr is passed, then all file_id[s] present in the
+            repository are returned
+        """
+        w = self._get_inventory_weave( )
+        if not last_revid:
+            changed = set(w._names)
+        else:
+            included = w.inclusions([w.lookup(last_revid)])
+            changed = map(w.idx_to_name, included)
+        return self._fileid_involved_by_set(changed)
+
+    def fileid_involved_by_set(self, changes):
+        """ This function returns the file_id(s) involved in the
+            changese present in the set changes
+        """
+        w = self._get_inventory_weave( )
+        return self._fileid_involved_by_set(changes)
+
+    def _fileid_involved_by_set(self, changes):
+        w = self._get_inventory_weave( )
+        file_ids = set( )
+        for line in w._weave:
+
+            # it is ugly, but it is due to the weave structure
+            if not isinstance(line,basestring): continue
+
+            start = line.find('file_id="')+9
+            if start < 9: continue
+            end = line.find('"',start)
+            assert end>= 0
+            file_id = line[start:end]
+
+            # check if file_id is already present
+            if file_id in file_ids: continue
+
+            start = line.find('revision="')+10
+            if start < 10: continue
+            end = line.find('"',start)
+            assert end>= 0
+            revision_id = line[start:end]
+
+            if revision_id in changes:
+                file_ids.add(file_id)
+
+        return file_ids
+
 
 class ScratchBranch(BzrBranch):
     """Special test class: a branch that cleans up after itself.

=== modified file 'bzrlib/clone.py'
--- bzrlib/clone.py	2005-11-29 01:12:50 +0000
+++ bzrlib/clone.py	2005-12-15 20:59:28 +0000
@@ -78,19 +78,15 @@
         note("basis_branch is not supported for fast weave copy yet.")
     branch_from.lock_read()
     try:
-        if not (branch_from.weave_store.listable()
-                and branch_from.revision_store.listable()):
-            return copy_branch_slower(branch_from, to_location, revision,
-                                      basis_branch)
         history = _get_truncated_history(branch_from, revision)
         if not bzrlib.osutils.lexists(to_location):
             os.mkdir(to_location)
         branch_to = Branch.initialize(to_location)
         mutter("copy branch from %s to %s", branch_from, branch_to)
         branch_to.working_tree().set_root_id(branch_from.get_root_id())
-        _copy_control_weaves(branch_from, branch_to)
-        _copy_text_weaves(branch_from, branch_to)
-        _copy_revision_store(branch_from, branch_to)
+        _copy_control_weaves(branch_from, branch_to, history)
+        _copy_text_weaves(branch_from, branch_to, history)
+        _copy_revision_store(branch_from, branch_to, history)
         branch_to.set_parent(branch_from.base)
         # must be done *after* history is copied across
         branch_to.append_revision(*history)
@@ -100,7 +96,6 @@
     finally:
         branch_from.unlock()
 
-
 def _get_truncated_history(branch_from, revision_id):
     history = branch_from.revision_history()
     if revision_id is None:
@@ -111,52 +106,19 @@
         raise InvalidRevisionId(revision_id=revision, branch=branch_from)
     return history[:idx+1]
 
-def _copy_text_weaves(branch_from, branch_to):
-    copy_all(branch_from.weave_store, branch_to.weave_store)
-
-
-def _copy_revision_store(branch_from, branch_to):
-    copy_all(branch_from.revision_store, branch_to.revision_store)
-
-
-def _copy_control_weaves(branch_from, branch_to):
+def _copy_text_weaves(branch_from, branch_to, history):
+    from_set = set(branch_from.get_ancestry(history[-1])[1:])
+    file_ids = branch_from.fileid_involved_by_set( from_set )
+    branch_to.weave_store.copy_multi(branch_from.weave_store, file_ids )
+
+def _copy_revision_store(branch_from, branch_to, history):
+    # copy all revision
+    from_set = set(branch_from.get_ancestry(history[-1])[1:])
+    branch_to.revision_store.copy_multi(branch_from.revision_store, from_set )
+
+def _copy_control_weaves(branch_from, branch_to, history):
     to_control = branch_to.control_weaves
     from_control = branch_from.control_weaves
+    # TODO Goffredo 20051215: we need only the minimal revision !!!!!
     to_control.copy_multi(from_control, ['inventory'])
 
-    
-def copy_branch_slower(branch_from, to_location, revision=None, basis_branch=None):
-    """Copy branch_from into the existing directory to_location.
-
-    revision
-        If not None, only revisions up to this point will be copied.
-        The head of the new branch will be that revision.  Must be a
-        revid or None.
-
-    to_location -- The destination directory; must either exist and be 
-        empty, or not exist, in which case it is created.
-
-    revno
-        The revision to copy up to
-
-    basis_branch
-        A local branch to copy revisions from, related to branch_from. 
-        This is used when branching from a remote (slow) branch, and we have
-        a local branch that might contain some relevant revisions.
-    """
-    assert isinstance(branch_from, Branch)
-    assert isinstance(to_location, basestring)
-    if not bzrlib.osutils.lexists(to_location):
-        os.mkdir(to_location)
-    br_to = Branch.initialize(to_location)
-    mutter("copy branch from %s to %s", branch_from, br_to)
-    if basis_branch is not None:
-        basis_branch.push_stores(br_to)
-    br_to.working_tree().set_root_id(branch_from.get_root_id())
-    if revision is None:
-        revision = branch_from.last_revision()
-    br_to.update_revisions(branch_from, stop_revision=revision)
-    build_working_dir(to_location)
-    br_to.set_parent(branch_from.base)
-    mutter("copied")
-    return br_to

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2005-12-05 03:24:19 +0000
+++ bzrlib/fetch.py	2005-12-15 20:59:28 +0000
@@ -108,23 +108,86 @@
             self.pb = pb
         self.from_branch.lock_read()
         try:
-            self._fetch_revisions(last_revision)
+            revs = self._revids_to_fetch(last_revision )
+            # nothing to do
+            if revs: 
+                self._fetch_revision_texts( revs )
+                self._fetch_weave_texts( revs )
+                self._fetch_inventory_weave( revs )
+                self.count_copied += len(revs)
         finally:
             self.from_branch.unlock()
             self.pb.clear()
 
-    def _fetch_revisions(self, last_revision):
+    def _revids_to_fetch(self, last_revision):
         self.last_revision = self._find_last_revision(last_revision)
         mutter('fetch up to rev {%s}', self.last_revision)
         if (self.last_revision is not None and 
             self.to_branch.has_revision(self.last_revision)):
             return
         try:
-            revs_to_fetch = self._compare_ancestries()
+            branch_from_revs = set(self.from_branch.get_ancestry(self.last_revision))
         except WeaveError:
             raise InstallFailed([self.last_revision])
-        self._copy_revisions(revs_to_fetch)
-        self.new_ancestry = revs_to_fetch
+
+        self.dest_last_rev = self.to_branch.last_revision()
+        branch_to_revs = set(self.to_branch.get_ancestry(self.dest_last_rev))
+
+        return branch_from_revs.difference( branch_to_revs )
+
+    def _fetch_revision_texts( self, revs ):
+        self.to_branch.revision_store.copy_multi(
+            self.from_branch.revision_store, revs )
+
+    def _fetch_weave_texts( self, revs ):
+        file_ids = self.from_branch.fileid_involved_by_set( revs )
+        count = 0
+        num_file_ids = len(file_ids)
+        for file_id in file_ids:
+            self.pb.update( "merge weave merge",count,num_file_ids)
+            count +=1
+            to_weave = self.to_weaves.get_weave_or_empty(file_id,
+                self.to_branch.get_transaction())
+            from_weave = self.from_weaves.get_weave(file_id,
+                self.from_branch.get_transaction())
+
+            if to_weave.numversions() > 0:
+                # destination has contents, must merge
+                try:
+                    to_weave.join(from_weave)
+                except errors.WeaveParentMismatch:
+                    to_weave.reweave(from_weave)
+            else:
+                # destination is empty, just replace it
+                to_weave = from_weave.copy( )
+
+            self.to_weaves.put_weave(file_id, to_weave,
+                self.to_branch.get_transaction())
+
+        self.pb.clear( )
+
+    def _fetch_inventory_weave( self, revs ):
+        self.pb.update( "inventory merge",0,1)
+
+        from_weave = self.from_control.get_weave('inventory',
+                self.from_branch.get_transaction())
+        to_weave = self.to_control.get_weave('inventory',
+                self.to_branch.get_transaction())
+
+        if to_weave.numversions() > 0:
+            # destination has contents, must merge
+            try:
+                to_weave.join(from_weave)
+            except errors.WeaveParentMismatch:
+                to_weave.reweave(from_weave)
+        else:
+            # destination is empty, just replace it
+            to_weave = from_weave.copy( )
+
+        self.to_control.put_weave('inventory', to_weave,
+            self.to_branch.get_transaction())
+
+        self.pb.clear( )
 
     def _find_last_revision(self, last_revision):
         """Find the limiting source revision.
@@ -142,113 +205,5 @@
             return from_history[-1]
         else:
             return None                 # no history in the source branch
-            
-
-    def _compare_ancestries(self):
-        """Get a list of revisions that must be copied.
-
-        That is, every revision that's in the ancestry of the source
-        branch and not in the destination branch."""
-        self.pb.update('get source ancestry')
-        self.from_ancestry = self.from_branch.get_ancestry(self.last_revision)
-
-        dest_last_rev = self.to_branch.last_revision()
-        self.pb.update('get destination ancestry')
-        if dest_last_rev:
-            dest_ancestry = self.to_branch.get_ancestry(dest_last_rev)
-        else:
-            dest_ancestry = []
-        ss = set(dest_ancestry)
-        to_fetch = []
-        for rev_id in self.from_ancestry:
-            if rev_id not in ss:
-                to_fetch.append(rev_id)
-                mutter('need to get revision {%s}', rev_id)
-        mutter('need to get %d revisions in total', len(to_fetch))
-        self.count_total = len(to_fetch)
-        return to_fetch
-
-    def _copy_revisions(self, revs_to_fetch):
-        i = 0
-        for rev_id in revs_to_fetch:
-            i += 1
-            if rev_id is None:
-                continue
-            if self.to_branch.has_revision(rev_id):
-                continue
-            self.pb.update('copy revision', i, self.count_total)
-            self._copy_one_revision(rev_id)
-            self.count_copied += 1
-
-
-    def _copy_one_revision(self, rev_id):
-        """Copy revision and everything referenced by it."""
-        mutter('copying revision {%s}', rev_id)
-        rev_xml = self.from_branch.get_revision_xml(rev_id)
-        inv_xml = self.from_branch.get_inventory_xml(rev_id)
-        rev = serializer_v5.read_revision_from_string(rev_xml)
-        inv = serializer_v5.read_inventory_from_string(inv_xml)
-        assert rev.revision_id == rev_id
-        assert rev.inventory_sha1 == sha_string(inv_xml)
-        mutter('  commiter %s, %d parents',
-               rev.committer,
-               len(rev.parent_ids))
-        self._copy_new_texts(rev_id, inv)
-        parents = rev.parent_ids
-        new_parents = copy(parents)
-        for parent in parents:
-            if not self.to_branch.has_revision(parent):
-                new_parents.pop(new_parents.index(parent))
-        self._copy_inventory(rev_id, inv_xml, new_parents)
-        self.to_branch.revision_store.add(StringIO(rev_xml), rev_id)
-        mutter('copied revision %s', rev_id)
-
-    def _copy_inventory(self, rev_id, inv_xml, parent_ids):
-        self.to_control.add_text('inventory', rev_id,
-                                split_lines(inv_xml), parent_ids,
-                                self.to_branch.get_transaction())
-
-    def _copy_new_texts(self, rev_id, inv):
-        """Copy any new texts occuring in this revision."""
-        # TODO: Rather than writing out weaves every time, hold them
-        # in memory until everything's done?  But this way is nicer
-        # if it's interrupted.
-        for path, ie in inv.iter_entries():
-            self._copy_one_weave(rev_id, ie.file_id, ie.revision)
-
-    def _copy_one_weave(self, rev_id, file_id, text_revision):
-        """Copy one file weave, esuring the result contains text_revision."""
-        # check if the revision is already there
-        if file_id in self.file_ids_names.keys( ) and \
-            text_revision in self.file_ids_names[file_id]:
-                return        
-        to_weave = self.to_weaves.get_weave_or_empty(file_id,
-            self.to_branch.get_transaction())
-        if not file_id in self.file_ids_names.keys( ):
-            self.file_ids_names[file_id] = to_weave.names( )
-        if text_revision in to_weave:
-            return
-        from_weave = self.from_weaves.get_weave(file_id,
-            self.from_branch.get_transaction())
-        if text_revision not in from_weave:
-            raise MissingText(self.from_branch, text_revision, file_id)
-        mutter('copy file {%s} modified in {%s}', file_id, rev_id)
-
-        if to_weave.numversions() > 0:
-            # destination has contents, must merge
-            try:
-                to_weave.join(from_weave)
-            except errors.WeaveParentMismatch:
-                to_weave.reweave(from_weave)
-        else:
-            # destination is empty, just replace it
-            to_weave = from_weave.copy( )
-        self.to_weaves.put_weave(file_id, to_weave,
-            self.to_branch.get_transaction())
-        self.count_weaves += 1
-        self.copied_file_ids.add(file_id)
-        self.file_ids_names[file_id] = to_weave.names()
-        mutter('copied file {%s}', file_id)
-
 
 fetch = Fetcher

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2006-01-06 01:13:05 +0000
+++ bzrlib/tests/__init__.py	2006-01-18 13:45:16 +0000
@@ -665,6 +665,7 @@
                    'bzrlib.tests.test_conflicts',
                    'bzrlib.tests.test_diff',
                    'bzrlib.tests.test_fetch',
+                   'bzrlib.tests.test_fileid_involved',
                    'bzrlib.tests.test_gpg',
                    'bzrlib.tests.test_graph',
                    'bzrlib.tests.test_hashcache',

=== modified file 'bzrlib/tests/test_fetch.py'
--- bzrlib/tests/test_fetch.py	2005-11-25 06:28:56 +0000
+++ bzrlib/tests/test_fetch.py	2005-12-15 19:34:35 +0000
@@ -86,7 +86,7 @@
         self.assertFalse(has_revision(br_a3, br_a.revision_history()[revno]))
     self.assertEqual(greedy_fetch(br_a3, br_a2, br_a.revision_history()[2])[0], 3)
     fetched = greedy_fetch(br_a3, br_a2, br_a.revision_history()[3])[0]
-    self.assertEquals(fetched, 3, "fetched %d instead of 3" % fetched)
+    self.assertEquals(fetched, 6, "fetched %d instead of 6" % fetched)
     # InstallFailed should be raised if the branch is missing the revision
     # that was requested.
     self.assertRaises(bzrlib.errors.InstallFailed, greedy_fetch, br_a3,

=== added file 'bzrlib/tests/test_fileid_involved.py'
--- bzrlib/tests/test_fileid_involved.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_fileid_involved.py	2005-12-15 21:37:04 +0000
@@ -0,0 +1,181 @@
+
+
+
+from bzrlib.tests import TestCaseInTempDir
+import os
+from bzrlib.commit import commit
+from bzrlib.add import smart_add
+from bzrlib.branch import Branch
+from bzrlib.clone import copy_branch
+from bzrlib.merge import merge
+from bzrlib.workingtree import WorkingTree
+from bzrlib.delta import compare_trees
+
+class TestFileIdInvolved(TestCaseInTempDir):
+
+    def touch(self,filename):
+        f = file(filename,"a")
+        f.write("appended line\n")
+        f.close( )
+
+
+    def merge( self, branch_from, force=False ):
+        from bzrlib.merge_core import ApplyMerge3
+
+        merge([branch_from,-1],[None,None], merge_type=ApplyMerge3,
+            check_clean=(not force) )
+
+    def setUp(self):
+        super(TestFileIdInvolved, self).setUp()
+        # create three branches, and merge it
+        #
+        #           /-->J ------>K                (branch2)
+        #          /              \
+        #  A ---> B --->C ---->D->G               (main)
+        #  \           /      /
+        #   \---> E---/----> F                 (branch1)
+
+        os.mkdir("main")
+        os.chdir("main")
+
+        main_branch = Branch.initialize('.')
+        self.build_tree(["a","b","c"])
+
+        smart_add('.')
+        commit(Branch.open("."), "Commit one", rev_id="rev-A")
+        #-------- end A -----------
+
+        copy_branch(main_branch,"../branch1")
+        os.chdir("../branch1")
+
+        #branch1_branch = Branch.open(".")
+        self.build_tree(["d"])
+        smart_add(".")
+        commit(Branch.open("."), "branch1, Commit one", rev_id="rev-E")
+
+        #-------- end E -----------
+
+        os.chdir("../main")
+        self.touch("a")
+        commit(Branch.open("."), "Commit two", rev_id="rev-B")
+
+        #-------- end B -----------
+
+        copy_branch(Branch.open("."),"../branch2")
+        os.chdir("../branch2")
+
+        branch2_branch = Branch.open(".")
+        os.chmod("b",0770)
+        commit(Branch.open("."), "branch2, Commit one", rev_id="rev-J")
+
+        #-------- end J -----------
+
+        os.chdir("../main")
+
+        self.merge("../branch1")
+        commit(Branch.open("."), "merge branch1, rev-11", rev_id="rev-C")
+
+        #-------- end C -----------
+
+        os.chdir("../branch1")
+        tree = WorkingTree('.', Branch.open("."))
+        tree.rename_one("d","e")
+        commit(Branch.open("."), "branch1, commit two", rev_id="rev-F")
+
+
+        #-------- end F -----------
+
+        os.chdir("../branch2")
+
+        self.touch("c")
+        smart_add('.')
+        commit(Branch.open("."), "branch2, commit two", rev_id="rev-K")
+
+        #-------- end K -----------
+
+        os.chdir("../main")
+
+        self.touch("b")
+        self.merge("../branch1",force=True)
+
+        commit(Branch.open("."), "merge branch1, rev-12", rev_id="rev-D")
+
+        # end D
+
+        self.merge("../branch2")
+        commit(Branch.open("."), "merge branch1, rev-22",  rev_id="rev-G")
+
+        # end G
+        os.chdir("../main")
+        self.branch = Branch.open(".")
+
+
+    def test_fileid_involved_all_revs(self):
+
+        l = self.branch.fileid_involved( )
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["a","b","c","d"])
+
+    def test_fileid_involved_one_rev(self):
+
+        l = self.branch.fileid_involved("rev-B" )
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["a","b","c"])
+
+    def test_fileid_involved_two_revs(self):
+
+        l = self.branch.fileid_involved_between_revs("rev-B","rev-K" )
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["b","c"])
+
+        l = self.branch.fileid_involved_between_revs("rev-C","rev-D" )
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["b","d"])
+
+        l = self.branch.fileid_involved_between_revs("rev-C","rev-G" )
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["b","c","d"])
+
+        l = self.branch.fileid_involved_between_revs("rev-E","rev-G" )
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["a", "b","c","d"])
+
+
+    def test_fileid_involved_sets(self):
+
+        l = self.branch.fileid_involved_by_set(set(["rev-B"]))
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["a"])
+
+        l = self.branch.fileid_involved_by_set(set(["rev-D"]))
+        self.assertEquals( sorted(map( lambda x: x[0], l )), ["b"])
+
+    def test_fileid_involved_compare(self):
+
+        l1 = self.branch.fileid_involved_between_revs("rev-E", "rev-D")
+        l2 = self.branch.fileid_involved_by_set(set(["rev-D","rev-F","rev-C","rev-B"]))
+        self.assertEquals( l1, l2 )
+
+        l1 = self.branch.fileid_involved_between_revs("rev-C", "rev-G")
+        l2 = self.branch.fileid_involved_by_set(
+            set(["rev-G","rev-D","rev-F","rev-K","rev-J"]))
+        self.assertEquals( l1, l2 )
+
+    def test_fileid_involved_full_compare(self):
+        from bzrlib.tsort import topo_sort
+        pp=[]
+        history = self.branch.revision_history( )
+
+        if len(history) < 2: return
+
+        for start in range(0,len(history)-1):
+            for end in range(start+1,len(history)):
+
+                l1 = self.branch.fileid_involved_between_revs(
+                    history[start], history[end])
+
+                old_tree = self.branch.revision_tree(history[start])
+                new_tree = self.branch.revision_tree(history[end])
+                delta = compare_trees(old_tree, new_tree )
+
+                l2 = [ id for path, id, kind in delta.added ] + \
+                     [ id for oldpath, newpath, id, kind, text_modified, \
+                            meta_modified in delta.renamed ] + \
+                     [ id for path, id, kind, text_modified, meta_modified in \
+                            delta.modified ]
+
+                self.assertEquals( l1, set(l2) )
+

